﻿using Dynamicweb.DataIntegration.EndpointManagement;
using Dynamicweb.DataIntegration.Integration.ERPIntegration;
using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Ecommerce.LiveIntegration.EndpointMonitoring;
using Dynamicweb.Ecommerce.LiveIntegration.Licensing;
using Dynamicweb.Ecommerce.LiveIntegration.Logging;
using Dynamicweb.Extensibility.Notifications;
using System;
using System.Xml;
using static Dynamicweb.Ecommerce.LiveIntegration.Notifications.Communication;

namespace Dynamicweb.Ecommerce.LiveIntegration
{
    /// <summary>
    /// Main class to interact with a remote ERP.
    /// </summary>
    internal static class Connector
    {
        /// <summary>
        /// The maximum retry count
        /// </summary>
        private static readonly int MaxRetryCount = 3;

        /// <summary>
        /// The retry interval
        /// </summary>
        private static readonly int RetryInterval = 500;

        public static bool EnableThrowExceptions
        {
            get
            {
                return Core.Converter.ToBoolean(Context.Current?.Items?["LiveIntegrationAddInThrowExceptions"]);
            }
            set
            {
                if (Context.Current?.Items != null)
                {
                    Context.Current.Items["LiveIntegrationAddInThrowExceptions"] = value;
                }
            }
        }

        /// <summary>
        /// Gets the security key.
        /// </summary>
        /// <value>The security key.</value>
        private static string SecurityKey => Settings.Instance.SecurityKey;

        /// <summary>
        /// Gets the URL.
        /// </summary>
        /// <value>The URL.</value>
        /// <exception cref="ArgumentException">
        /// Setup does not contain a url for the WebService. Please rerun setup.
        /// or
        /// Cannot find appropriate web service url. Check Web Service Url settings.
        /// </exception>
        private static string Url
        {
            get
            {
                string url = UrlHandler.Instance.GetWebServiceUrl();
                if (!string.IsNullOrEmpty(url))
                {
                    return url;
                }

                if (string.IsNullOrEmpty(Settings.Instance.WebServiceURI))
                {
                    throw new ArgumentException("Setup does not contain a url for the WebService. Please rerun setup.");
                }
                else
                {
                    throw new ArgumentException("Cannot find appropriate web service url. Check Web Service Url settings.");
                }
            }
        }

        /// <summary>
        /// Calculates the order.
        /// </summary>
        /// <param name="orderXml">The order XML.</param>
        /// <param name="orderId">The order identifier.</param>
        /// <param name="createOrder">if set to <c>true</c> [create order].</param>
        /// <param name="error">The error.</param>
        /// <returns>XmlDocument.</returns>
        public static XmlDocument CalculateOrder(string orderXml, string orderId, bool createOrder, out Exception error)
        {
            error = null;
            XmlDocument document = null;

            try
            {
                // only retry if is not create order, for create order schedule task will send later
                document = Communicate(orderXml, $"CalculateOrder (ID: {orderId}, CreateOrder: {createOrder})", !createOrder, true);
            }
            catch (Exception ex)
            {
                error = ex;
            }

            if (EnableThrowExceptions && error != null)
            {
                throw error;
            }

            return document;
        }

        /// <summary>
        /// Clears the cache.
        /// </summary>
        public static void ClearCache()
        {
            EndpointMonitoringService.ClearEndpoints();
        }

        /// <summary>
        /// Gets the products information.
        /// </summary>
        /// <param name="productsXml">The products XML.</param>
        /// <param name="retry">Retry request on error.</param>
        /// <returns>XmlDocument.</returns>
        public static XmlDocument GetProductsInfo(string productsXml, bool retry = false)
        {
            return Communicate(productsXml, "GetProductsInfo", retry);
        }

        /// <summary>
        /// Determines whether the web service connection is available.
        /// </summary>
        public static bool IsWebServiceConnectionAvailable()
        {
            return IsWebServiceConnectionAvailable(null);
        }

        /// <summary>
        /// Determines whether the web service connection is available.
        /// </summary>
        public static bool IsWebServiceConnectionAvailable(string url)
        {
            if (!Settings.Instance.IsLiveIntegrationEnabled)
            {
                return false;
            }

            if (string.IsNullOrEmpty(url))
            {
                url = Url;
                bool connectionAvailable = false;
                if (EndpointMonitoringService.IsStillValid(url, out connectionAvailable))
                {
                    return connectionAvailable;
                }
            }

            var lastStatus = EndpointMonitoringService.GetEndpointStatus(url);
            bool isWebServiceConnectionAvailable = (lastStatus != null) ? !lastStatus.Error : false;
            DateTime? lastErpCommunication = (lastStatus != null) ? lastStatus.LastErpCommunication : null;
            bool success = false;

            string request = "<GetEcomData></GetEcomData>"; // simple request for checking connection
            string response = null;
            try
            {
                TimeSpan ts = TimeSpan.Zero;
                if(Settings.Instance.ConnectionTimeout > 0)
                {
                    ts = new TimeSpan(0, 0, Settings.Instance.ConnectionTimeout);
                }
                response = Execute(url, request, ts);
                success = true;
                EndpointMonitoringService.Success(url);
            }

            catch (Exception ex)
            {
                Logger.Instance.Log(ErrorLevel.ConnectionError, $"Error checking Web Service connection: '{ex.Message}'. Request: '{request}'.");

                EndpointMonitoringService.Error(url);

                NotificationManager.Notify(OnErpCommunicationLost, new OnErpCommunicationLostArgs(request, ex));

                if (EnableThrowExceptions)
                {
                    throw new Exception("Thrown exception", ex);
                }
            }
            if (success && lastErpCommunication.HasValue && !isWebServiceConnectionAvailable)
            {
                NotificationManager.Notify(OnErpCommunicationRestored, new OnErpCommunicationRestoredArgs(request, lastErpCommunication));
            }
            if(success && !string.IsNullOrEmpty(response))
            {
                LicenseService.ValidateLicense(url, response);
            }
            return success;
        }

        /// <summary>
        /// Retrieves data from the request string.
        /// </summary>
        /// <param name="request">The request.</param>
        /// <returns>XmlDocument.</returns>
        public static XmlDocument RetrieveDataFromRequestString(string request)
        {
            return Communicate(request, "RetrieveDataFromRequestString");
        }

        /// <summary>
        /// Communicates the specified request.
        /// </summary>
        /// <param name="request">The request.</param>
        /// <param name="referenceName">Name of the reference.</param>
        /// <param name="retry">if set to <c>true</c> [retry].</param>
        /// <param name="throwException">if set to <c>true</c> [throw exception].</param>
        /// <returns>XmlDocument.</returns>
        private static XmlDocument Communicate(string request, string referenceName, bool retry = true, bool throwException = false)
        {
            XmlDocument result = null;
            int retryCount = 0;

            // if ERP is down and in retry mode wait!
            while (!IsWebServiceConnectionAvailable())
            {
                if (!retry || retryCount >= MaxRetryCount)
                {
                    return null;
                }

                ++retryCount;
                System.Threading.Thread.Sleep(RetryInterval);
            }

            if (!LicenseService.IsLicenseValid(Url))
            {
                Logger.Instance.Log(ErrorLevel.DebugInfo, $"Request send failed. License is not valid.");
                return null;
            }

            Exception exception = null;
            do
            {
                exception = null;
                string erpXmlResponse = null;
                try
                {
                    NotificationManager.Notify(OnBeforeErpCommunication, new OnBeforeErpCommunicationArgs(request, referenceName));

                    Logger.Instance.Log(ErrorLevel.DebugInfo, $"Request {referenceName} sent: '{request}'.");

                    erpXmlResponse = Execute(Url, request);

                    EndpointMonitoringService.Success(Url);

                    if (!string.IsNullOrEmpty(erpXmlResponse))
                    {
                        if (!Helpers.ParseResponseToXml(erpXmlResponse, out result))
                        {
                            result = null;
                            exception = new Exception("Response is not valid XML");
                        }

                        if (!LicenseService.ValidateLicense(Url, result))
                        {
                            Logger.Instance.Log(ErrorLevel.DebugInfo, $"Response failed. License is not valid.");
                            result = null;
                            retry = false;
                        }
                        Logger.Instance.Log(ErrorLevel.DebugInfo, $"Response {referenceName} received: '{erpXmlResponse}'.");
                    }
                    else
                    {
                        Logger.Instance.Log(ErrorLevel.ResponseError, $"Response {referenceName} returned null.");
                    }
                }
                catch (Exception ex)
                {
                    exception = ex;

                    Logger.Instance.Log(ErrorLevel.ConnectionError, $"An error occurred while calling {referenceName} from Web Service: '{ex.Message}'.");

                    EndpointMonitoringService.Error(Url);

                    NotificationManager.Notify(OnAfterErpException, new OnAfterErpExceptionArgs(request, erpXmlResponse, referenceName, ex));
                }
                finally
                {
                    NotificationManager.Notify(OnAfterErpCommunication, new OnAfterErpCommunicationArgs(request, erpXmlResponse, referenceName, exception));
                }

                // no retry or xml reply from ERP then leave
                if (!retry || result != null)
                {
                    break;
                }

                // in retry mode, wait some time
                ++retryCount;
                System.Threading.Thread.Sleep(RetryInterval);
            }
            while (retryCount < MaxRetryCount);

            if ((throwException || EnableThrowExceptions) && exception != null)
            {
                throw exception;
            }
            return result;
        }

        internal static string RetrievePDF(string requestString)
        {
            string base64EncodedPDF = string.Empty;
            try
            {
                Logger.Instance.Log(ErrorLevel.DebugInfo, string.Format("Request RetrievePDF sent: '{0}'.", requestString));
                base64EncodedPDF = Execute(UrlHandler.Instance.GetWebServiceUrl(), requestString);
                Logger.Instance.Log(ErrorLevel.DebugInfo, string.Format("Response RetrievePDF received: '{0}'.", base64EncodedPDF));
            }
            catch (Exception ex)
            {
                Logger.Instance.Log(ErrorLevel.ResponseError, string.Format("Response RetrievePDF returned error: '{0}'.", ex.Message));
                throw ex;
            }
            return base64EncodedPDF;
        }

        private static string Execute(string url, string request)
        {
            return Execute(url, request, TimeSpan.Zero);
        }
        private static string Execute(string url, string request, TimeSpan responseTimeout)
        {
            string response = null;
            int endpointId = Settings.Instance.EndpointId;
            if (endpointId > 0)
            {
                EndpointService endpointService = new EndpointService();
                var endpoint = endpointService.GetEndpointById(endpointId);
                if(endpoint != null)
                {
                    response = endpointService.Execute(endpointId, endpointService.GetDynamicwebServiceSoapBody(request));
                }
                else
                {
                    Logger.Instance.Log(ErrorLevel.DebugInfo, $"Endpoint {endpointId} does not exist.");
                }
            }
            else
            {
                response = ErpServiceCaller.GetDataFromRequestString(url, SecurityKey, request, responseTimeout);
            }
            return response;
        }
    }
}