﻿using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Ecommerce.LiveIntegration.Extensions;
using Dynamicweb.Ecommerce.LiveIntegration.Logging;
using Dynamicweb.Ecommerce.LiveIntegration.Products;
using Dynamicweb.Ecommerce.Products;
using Dynamicweb.Extensibility.Notifications;
using Dynamicweb.Security.UserManagement;
using System;
using System.Collections.Generic;
using System.Linq;

namespace Dynamicweb.Ecommerce.LiveIntegration.NotificationSubscribers
{
    /// <summary>
    /// Base notification subscriber for Live Integration, for common functionality.
    /// </summary>
    /// <seealso cref="NotificationSubscriber" />
    public abstract class NotificationSubscriberBase : NotificationSubscriber
    {
        /// <summary>
        /// Gets a value indicating whether prices can be looked up in the ERP. 
        /// </summary>
        protected bool CanCheckPrice
        {
            get
            {
                return Settings.Instance.EnableLivePrices && Global.IsIntegrationActive && (Settings.Instance.LiveProductInfoForAnonymousUsers || User.GetCurrentExtranetUser() != null) && Global.IsIntegrationEnabledForShop && (User.GetCurrentExtranetUser() == null || !User.GetCurrentExtranetUser().IsLivePricesDisabled) && Connector.IsWebServiceConnectionAvailable() && !Global.IsProductLazyLoad;
            }
        }

        /// <summary>
        /// Updates the product information.
        /// </summary>
        /// <param name="order">The order.</param>
        internal static void UpdateProductInformation(Orders.Order order)
        {
            if (order == null || !order.OrderLines.Any(ol => ol.IsProduct()))
            {
                // nothing to update
                return;
            }

            // clear product cache to ensure refresh from ERP
            Integration.ErpResponseCache.ClearAllCaches();
            ProductManager.ClearUserCache();

            // read all products in the order
            var productsToUpdate = GetProductsWithQuantities(order);
            if (productsToUpdate.Any())
            {
                Logger.Instance.Log(ErrorLevel.DebugInfo, $"Reload prices. products #{productsToUpdate.Count()}");
                var orderUser = User.GetUserByID(order.CustomerAccessUserId);
                SetProductsInfo(productsToUpdate, orderUser);
            }
        }

        /// <summary>
        /// Checks the order prices.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="notification">The notification.</param>
        protected static void CheckOrderPrices(Orders.Order order, string notification)
        {
            if (order == null || order.OrderLines.Count <= 0 || order.Complete)
            {
                return;
            }

            Dictionary<Product, double> products = GetProductsWithQuantities(order);

            if (ProductManager.FetchProductInfos(products, null))
            {
                foreach (var productWithQuantity in products)
                {
                    var product = productWithQuantity.Key;
                    ProductManager.FillProductValues(product, productWithQuantity.Value);

                    var orderLine = order.OrderLines.First(ol => string.Compare(ol.ProductId, product.Id, StringComparison.OrdinalIgnoreCase) == 0);

                    if (!Helpers.IsDifferentPrice(product.Price.PriceWithVAT, orderLine.UnitPrice.PriceWithVAT) && !Helpers.IsDifferentPrice(product.Price.PriceWithoutVAT, orderLine.UnitPrice.PriceWithoutVAT))
                    {
                        if (!Helpers.IsDifferentPrice(orderLine.Price.PriceWithVAT, orderLine.Quantity * orderLine.UnitPrice.PriceWithVAT) && !Helpers.IsDifferentPrice(orderLine.Price.PriceWithoutVAT, orderLine.Quantity * orderLine.UnitPrice.PriceWithoutVAT))
                        {
                            continue;
                        }
                    }
                    
                    orderLine.AllowOverridePrices = true;                    
                    Services.OrderLines.SetUnitPrice(orderLine, product.Price, false);
                    orderLine.OrderLineType = Orders.OrderLineType.Fixed;
                    orderLine.Price.PriceWithVAT = orderLine.Quantity * orderLine.UnitPrice.PriceWithVAT;
                    orderLine.Price.PriceWithoutVAT = orderLine.Quantity * orderLine.UnitPrice.PriceWithoutVAT;
                }
            }

            Services.Orders.ForcePriceRecalculation(order);
        }
        
        /// <summary>
        /// Gets a value that determines whether live integration is enabled for the current shop, active and a connection to the ERP is available.
        /// </summary>
        /// <returns><c>true</c> if the live integration is enabled and active, <c>false</c> otherwise.</returns>
        protected static bool EnabledAndActive()
        {
            return Global.IsIntegrationActive
              && Global.IsIntegrationEnabledForShop
              && Connector.IsWebServiceConnectionAvailable();
        }

        /// <summary>
        /// Gets the products and their current quantities from the order lines.
        /// </summary>
        /// <param name="order">The order from which the products are retrieved.</param>
        protected static Dictionary<Product, double> GetProductsWithQuantities(Orders.Order order)
        {
            Dictionary<Product, double> products = new Dictionary<Product, double>();
            foreach (var ol in order.OrderLines)
            {
                if (ol.IsProduct() && !products.ContainsKey(ol.Product))
                {
                    products.Add(ol.Product, ol.Quantity);
                }
            }
            return products;
        }

        /// <summary>
        /// Sets the product information in Dynamicweb from the ERP.
        /// </summary>
        /// <param name="product">The product to be updated.</param>
        /// <param name="quantity">The quantity.</param>
        protected void SetProductInformation(Product product, double quantity)
        {
            SetProductsInfo(new Dictionary<Product, double> { { product, quantity } });
        }

        /// <summary>
        /// Sets the product information in Dynamicweb for all the products passed.
        /// </summary>
        /// <param name="products">The products to be updated.</param>
        /// <param name="user">The user.</param>
        protected void SetProductInformation(Dictionary<Product, double> products, User user = null)
        {
            SetProductsInfo(products, user);
        }

        /// <summary>
        /// Clears and sets the product information in Dynamicweb for all the products in the order.
        /// </summary>
        /// <param name="order">The order.</param>
        protected void SetProductInformation(Orders.Order order)
        {
            if (order == null)
            {
                return;
            }

            // clear product cache to ensure refresh from ERP
            Integration.ErpResponseCache.ClearAllCaches();
            ProductManager.ClearUserCache();

            // read all products in the order
            var productsToUpdate = GetProductsWithQuantities(order);
            if (productsToUpdate != null && productsToUpdate.Any())
            {
                Logger.Instance.Log(ErrorLevel.DebugInfo, $"Reload prices. products #{productsToUpdate.Count()}");
                User orderUser = User.GetUserByID(order.CustomerAccessUserId);
                SetProductsInfo(productsToUpdate, orderUser);
            }
        }

        /// <summary>
        /// Updates the order.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="notification">The notification.</param>
        protected void UpdateOrder(Orders.Order order, string notification)
        {
            if (order != null)
            {
                if (Global.EnableCartCommunication(order.Complete))
                {
                    OrderHandler.UpdateOrder(order, SubmitType.LiveOrderOrCart);
                }
                else
                {
                    try
                    {
                        CheckOrderPrices(order, notification);
                    }
                    catch (Exception ex)
                    {
                        string message = $"Error updating order prices on {notification}. order# {order.Id}. Error:{ex}.";
                        Logger.Instance.Log(ErrorLevel.Error, message);
                        if (Connector.EnableThrowExceptions)
                        {
                            throw new Exception(message, ex);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Sets the products information.
        /// </summary>
        /// <param name="products">The products.</param>
        /// <param name="user">The user.</param>
        private static void SetProductsInfo(Dictionary<Product, double> products, User user = null)
        {
            if (!ProductManager.FetchProductInfos(products, user))
            {
                return;
            }

            // Set values
            foreach (var productWithQuantity in products)
            {
                ProductManager.FillProductValues(productWithQuantity.Key, productWithQuantity.Value);
            }
        }
    }
}