﻿using Dynamicweb.Caching;
using Dynamicweb.Ecommerce.Integration;
using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Ecommerce.LiveIntegration.Logging;
using Dynamicweb.Ecommerce.LiveIntegration.XmlGenerators;
using Dynamicweb.Ecommerce.Prices;
using Dynamicweb.Ecommerce.Products;
using Dynamicweb.Ecommerce.Variants;
using Dynamicweb.Extensibility.AddIns;
using Dynamicweb.Frontend;
using Dynamicweb.Security.UserManagement;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;

namespace Dynamicweb.Ecommerce.LiveIntegration.Products
{
    /// <summary>
    /// Static manager to handle product info requests.
    /// </summary>
    public static class ProductManager
    {
        /// <summary>
        /// The fetch product info page cache
        /// </summary>
        private static readonly Dictionary<int, FetchProductInfoResult> FetchProductInfoPageCache = new Dictionary<int, FetchProductInfoResult>();

        private static readonly string ProductProviderCacheKey = "LiveIntegrationProductProvider";

        /// <summary>
        /// Gets the product cache level.
        /// </summary>
        /// <value>The product cache level.</value>
        public static CacheLevel ProductCacheLevel
        {
            get
            {
                string cacheLevelString = Settings.Instance.ProductCacheLevel;
                return Helpers.GetEnumValueFromString(cacheLevelString, CacheLevel.Page);
            }
        }

        /// <summary>
        /// Gets the product provider. Implement your own class deriving ProductProviderBase if you want to override the built-in behavior.
        /// </summary>
        /// <value>The product provider.</value>
        internal static ProductProviderBase ProductProvider
        {
            get
            {
                ProductProviderBase provider = null;

                if (!Cache.Current.TryGet(ProductProviderCacheKey, out provider))
                {
                    foreach (Type addIn in AddInManager.GetAddInClasses(typeof(ProductProviderBase)))
                    {
                        if (!ReferenceEquals(addIn, typeof(ProductProviderBase)))
                        {
                            provider = (ProductProviderBase)AddInManager.CreateAddInInstance(addIn);
                            break;
                        }
                    }

                    if (provider == null)
                    {
                        provider = new ProductProviderBase();
                    }

                    Cache.Current[ProductProviderCacheKey] = provider;
                }

                return provider;
            }
        }

        /// <summary>
        /// Fetches the product info from the ERP.
        /// </summary>
        /// <param name="products">The products.</param>
        /// <returns><c>true</c> if products information was fetched, <c>false</c> otherwise.</returns>
        public static bool FetchProductInfos(List<Product> products)
        {
            return FetchProductInfos(GetProductsWithQuantities(products), null);
        }

        /// <summary>
        /// Fills the product values from the response.
        /// </summary>
        /// <param name="product">The product.</param>
        /// <param name="quantity">The quantity.</param>
        public static void FillProductValues(Product product, double quantity)
        {
            ProductInfo productInfo = GetProductInfo(product);
            if (productInfo != null)
            {
                ProductProvider.FillProductValues(productInfo, product, quantity);
            }
        }

        /// <summary>
        /// Gets the product prices list from the response.
        /// </summary>
        /// <param name="product">The product.</param>        
        public static List<Price> GetProductPrices(Product product)
        {
            List<Price> prices = new List<Price>();
            ProductInfo productInfo = GetProductInfo(product);
            if (productInfo != null)
            {
                prices = ProductProvider.GetProductPrices(productInfo);
            }
            return prices;
        }        

        /// <summary>
        /// Gets the products and their current quantities.
        /// </summary>
        /// <param name="products"></param>
        /// <returns></returns>
        public static Dictionary<Product, double> GetProductsWithQuantities(IEnumerable<Product> products)
        {
            Dictionary<Product, double> productsWithQuantities = new Dictionary<Product, double>();
            foreach (var product in products)
            {
                if (!productsWithQuantities.ContainsKey(product))
                {
                    productsWithQuantities.Add(product, 1);
                }
            }

            return productsWithQuantities;
        }

        /// <summary>
        /// Clears the user cache.
        /// </summary>
        internal static void ClearUserCache()
        {
            if (ProductCacheLevel == CacheLevel.Session)
            {
                var cacheKey = GetCurrentUserProductInfoCacheKey();
                Context.Current.Session.Remove(cacheKey);
            }
            else
            {
                FetchProductInfoPageCache.Clear();
            }
        }

        /// <summary>
        /// Fetches the product info from the ERP.
        /// </summary>
        /// <param name="products">The products with quantities.</param>
        /// <param name="user">The user.</param>
        /// <param name="updateCache">Update cache with products information retrieved.</param>
        /// <returns><c>true</c> if product information was fetched, <c>false</c> otherwise.</returns>
        internal static bool FetchProductInfos(Dictionary<Product, double> products, User user, bool updateCache = true)
        {
            if (IsLastResponseValid())
            {
                // No need to read cache
                return false;
            }

            SaveLastResponse(false);

            if (products == null || products.Count == 0)
            {
                return false;
            }

            if (!Settings.Instance.EnableLivePrices)
            {
                Logger.Instance.Log(ErrorLevel.DebugInfo, "Live Prices are Disabled");
                return false;
            }

            if (!Connector.IsWebServiceConnectionAvailable())
            {
                Logger.Instance.Log(ErrorLevel.DebugInfo, "Live Prices are unavailable");
                return false;
            }

            if (user == null)
            {
                user = User.GetCurrentExtranetUser();
            }

            if (user == null && !Settings.Instance.LiveProductInfoForAnonymousUsers)
            {
                Logger.Instance.Log(ErrorLevel.DebugInfo, "No user is currently logged in. Anonymous user is not allowed.");
                return false;
            }

            if (user != null && user.IsLivePricesDisabled)
            {
                Logger.Instance.Log(ErrorLevel.DebugInfo, $"Calculated prices are not allowed for the user '{user.UserName}'.");
                return false;
            }

            // Check for existence of the given products in the Cache
            Dictionary<Product, double> productsForRequest = GetProductsForRequest(products);

            if (productsForRequest.Count == 0)
            {
                SaveLastResponse();
                return true;
            }

            string request = BuildProductRequest(productsForRequest);

            XmlDocument response = Connector.GetProductsInfo(request, Settings.Instance.MakeRetryForLiveProductInformation);

            if (response != null && !string.IsNullOrEmpty(response.InnerXml))
            {
                // Parse the response
                Dictionary<string, ProductInfo> prices = ProcessResponse(response);

                if (prices != null && prices.Count > 0)
                {
                    if (updateCache)
                    {
                        // Cache prices
                        foreach (string productKey in prices.Keys)
                        {
                            Dictionary<string, ProductInfo> cachedProductInfos = ErpResponseCache.GetProductInfos(ProductCacheLevel);

                            if (cachedProductInfos.ContainsKey(productKey))
                            {
                                cachedProductInfos.Remove(productKey);
                            }

                            cachedProductInfos.Add(productKey, prices[productKey]);
                        }
                    }
                }
                else
                {
                    return false;
                }
            }
            else
            {
                // error occurred
                return false;
            }

            SaveLastResponse();

            return true;
        }

        /// <summary>
        /// Builds the product request.
        /// </summary>
        /// <param name="products">The products.</param>
        /// <returns>System.String.</returns>
        private static string BuildProductRequest(Dictionary<Product, double> products)
        {
            var xmlGenerator = new ProductInfoXmlGenerator();
            var xmlGeneratorSettings = new ProductInfoXmlGeneratorSettings
            {
                AddProductFieldsToRequest = Settings.Instance.AddProductFieldsToRequest,
                Beautify = false,
                LiveIntegrationSubmitType = SubmitType.Live,
                ReferenceName = "ProductInfoLive"
            };

            var xml = xmlGenerator.GenerateProductInfoXml(
                products,
                xmlGeneratorSettings);

            return xml;
        }

        /// <summary>
        /// Session the cache key.
        /// </summary>
        /// <returns>System.String.</returns>
        private static string GetCurrentUserProductInfoCacheKey()
        {
            var user = User.GetCurrentExtranetUser();
            var cacheKey = Constants.CacheConfiguration.FetchProductInfoResult;

            if (user != null)
            {
                string userIdentifier = Helpers.GetCustomerIdentifier(user);
                if (string.IsNullOrEmpty(userIdentifier))
                {
                    userIdentifier = user.ID.ToString();
                }

                cacheKey += userIdentifier;
            }
            else
            {
                cacheKey += Settings.Instance.AnonymousUserKey;
            }

            return cacheKey;
        }

        private static Dictionary<Product, double> GetProductsForRequest(Dictionary<Product, double> products)
        {
            // Check for existence of the given products in the Cache
            Dictionary<Product, double> productsForRequest = new Dictionary<Product, double>();
            foreach (var productWithQuantity in products)
            {
                Product product = productWithQuantity.Key;
                string productIdentifier = ProductProvider.GetProductIdentifier(product);
                bool isProductCached = ErpResponseCache.IsProductInCache(ProductCacheLevel, productIdentifier);

                if (!isProductCached)
                {
                    // setting to request all variants in request
                    if (Settings.Instance.GetProductInformationForAllVariants)
                    {
                        var variants = VariantCombination.GetProductCombinations(product)?.Select(vc => vc.Product).ToList();
                        if (variants != null)
                        {
                            foreach (var variant in variants)
                            {
                                var variantProductIdentifier = ProductProvider.GetProductIdentifier(variant);
                                var isVariantProductCached = ErpResponseCache.IsProductInCache(ProductCacheLevel, variantProductIdentifier);

                                if (!isVariantProductCached)
                                {
                                    var isVariantProductAdded = productsForRequest.ContainsKey(variant);
                                    var isVariantProductLivePriceEnabled = ProductProvider.IsLivePriceEnabledForProduct(variant);

                                    if (!isVariantProductAdded && isVariantProductLivePriceEnabled)
                                    {
                                        productsForRequest.Add(variant, productWithQuantity.Value);
                                    }
                                }
                            }
                        }
                    }

                    product = ProductProvider.GetProductFromVariantComboId(product);

                    bool isProductAdded = productsForRequest.ContainsKey(product);
                    bool isProductLivePriceEnabled = ProductProvider.IsLivePriceEnabledForProduct(product);

                    if (!isProductAdded && isProductLivePriceEnabled)
                    {
                        productsForRequest.Add(product, productWithQuantity.Value);
                    }
                }
            }

            return productsForRequest;
        }

        /// <summary>
        /// Checks if the lasts response is valid.
        /// </summary>
        /// <returns><c>true</c> if last response is valid, <c>false</c> otherwise.</returns>
        private static bool IsLastResponseValid()
        {
            FetchProductInfoResult fetchResult = null;
            if (ProductCacheLevel == CacheLevel.Session)
            {
                var cacheKey = GetCurrentUserProductInfoCacheKey();
                object value = Context.Current.Session[cacheKey];

                if (value != null)
                {
                    fetchResult = value as FetchProductInfoResult;
                }
            }
            else
            {
                var pageId = PageView.Current().ID;
                if (FetchProductInfoPageCache.ContainsKey(pageId))
                {
                    fetchResult = FetchProductInfoPageCache[pageId];
                }
            }

            if (fetchResult != null && !fetchResult.Response && fetchResult.Executed.Subtract(DateTime.Now).Minutes < 5)
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Processes the response with product info.
        /// </summary>
        /// <param name="response">The response.</param>
        /// <returns>Dictionary&lt;System.String, ProductInfo&gt;.</returns>
        private static Dictionary<string, ProductInfo> ProcessResponse(XmlDocument response)
        {
            if (response == null)
            {
                return null;
            }

            try
            {
                Dictionary<string, ProductInfo> infos = new Dictionary<string, ProductInfo>();
                Func<string, string, string> priceLookupKey = (productId, variantId) => string.Join("|", productId, variantId);
                ILookup<string, ProductPrice> pricesLookup = ProductProvider.ExtractPrices(response).ToLookup(p => priceLookupKey(p.ProductId, p.ProductVariantId), StringComparer.OrdinalIgnoreCase);

                var items = response.SelectNodes("//item [@table='EcomProducts']");
                if (items != null)
                {
                    foreach (XmlNode item in items)
                    {
                        string productId = item.SelectSingleNode("column [@columnName='ProductId']")?.InnerText ?? string.Empty;
                        string variantId = item.SelectSingleNode("column [@columnName='ProductVariantId']")?.InnerText ?? string.Empty;
                        string productIdentifier = item.SelectSingleNode("column [@columnName='ProductIdentifier']")?.InnerText ?? string.Empty;
                        string productPrice = item.SelectSingleNode("column [@columnName='ProductPrice']")?.InnerText ?? string.Empty;

                        ProductInfo productInfo = new ProductInfo();

                        // Set ProductId
                        productInfo["ProductId"] = productIdentifier;

                        // Set price
                        productInfo["TotalPrice"] = !string.IsNullOrEmpty(productPrice) ? Helpers.ToDouble(productPrice) : (double?)null;

                        // Set stock
                        var stock = item.SelectSingleNode("column [@columnName='ProductStock']")?.InnerText;
                        if (stock != null)
                        {
                            productInfo["Stock"] = Helpers.ToDouble(stock);
                        }
                        else
                        {
                            productInfo["Stock"] = null;
                        }

                        // Set prices
                        productInfo["Prices"] = pricesLookup[priceLookupKey(productId, variantId)].ToList();

                        // Set ProductCustomFields
                        if (Settings.Instance.AddProductFieldsToRequest)
                        {
                            foreach (ProductField pf in ProductField.GetProductFields())
                            {
                                var fieldNode = item.SelectSingleNode($"column [@columnName='{pf.SystemName}']");
                                if (fieldNode != null)
                                {
                                    productInfo[pf.SystemName] = fieldNode.InnerText;
                                }
                            }
                        }

                        // avoid exception to duplicate products in XML
                        if (!infos.ContainsKey(productIdentifier))
                        {
                            infos.Add(productIdentifier, productInfo);
                        }
                    }
                }

                return infos;
            }
            catch (Exception e)
            {
                Logger.Instance.Log(ErrorLevel.Error, $"Response does not match schema: '{e.Message}'.");
            }

            return null;
        }

        /// <summary>
        /// Saves the last response.
        /// </summary>
        /// <param name="fetchResponse">if set to <c>true</c> [fetch response].</param>
        private static void SaveLastResponse(bool fetchResponse = true)
        {
            FetchProductInfoResult fetchResult = new FetchProductInfoResult
            {
                Response = fetchResponse
            };

            if (ProductCacheLevel == CacheLevel.Session)
            {
                var cacheKey = GetCurrentUserProductInfoCacheKey();
                Context.Current.Session[cacheKey] = fetchResult;
            }
            else
            {
                var pageId = PageView.Current().ID;
                if (FetchProductInfoPageCache.ContainsKey(pageId))
                {
                    FetchProductInfoPageCache[pageId] = fetchResult;
                }
                else
                {
                    FetchProductInfoPageCache.Add(pageId, fetchResult);
                }
            }
        }

        private static ProductInfo GetProductInfo(Product product)
        {
            ProductInfo productInfo = null;
            if (product != null)
            {
                string productIdentifier = ProductProvider.GetProductIdentifier(product);
                if (ErpResponseCache.IsProductInCache(ProductCacheLevel, productIdentifier))
                {
                    Dictionary<string, ProductInfo> productInfoCache = ErpResponseCache.GetProductInfos(ProductCacheLevel);
                    productInfoCache.TryGetValue(productIdentifier, out productInfo);
                }
            }
            return productInfo;
        }
    }
}