﻿using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Ecommerce.LiveIntegration.Extensions;
using Dynamicweb.Ecommerce.LiveIntegration.Products;
using Dynamicweb.Ecommerce.Products;
using Dynamicweb.Extensibility.Notifications;
using Dynamicweb.Security.UserManagement;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;

namespace Dynamicweb.Ecommerce.LiveIntegration.XmlGenerators
{
    /// <summary>
    /// Generates XML for product info requests.
    /// </summary>
    /// <seealso cref="XmlGenerator" />
    internal class ProductInfoXmlGenerator : XmlGenerator
    {
        /// <summary>
        /// Generates the product information XML.
        /// </summary>
        /// <param name="products">The products.</param>
        /// <param name="settings">The settings.</param>
        /// <returns>System.String.</returns>
        internal string GenerateProductInfoXml(Dictionary<Product, double> products, ProductInfoXmlGeneratorSettings settings)
        {
            NotificationManager.Notify(Notifications.ProductInfo.OnBeforeGenerateProductInfoXml, new Notifications.ProductInfo.OnBeforeGenerateProductInfoXmlArgs(products, settings));
            var xmlDocument = BuildXmlDocument();

            var xmlRoot = xmlDocument.CreateElement("GetEcomData");
            var user = User.GetCurrentExtranetUser();
            xmlRoot.SetAttribute("ExternalUserId", !string.IsNullOrWhiteSpace(user?.ExternalID) ? user.ExternalID : Settings.Instance.AnonymousUserKey);
            xmlRoot.SetAttribute("AccessUserCustomerNumber", !string.IsNullOrWhiteSpace(user?.CustomerNumber) ? user.CustomerNumber : Settings.Instance.AnonymousUserKey);

            var tablesNode = xmlDocument.CreateElement("tables");
            tablesNode.AppendChild(BuildProductInfoXml(xmlDocument, products, settings));

            xmlRoot.AppendChild(tablesNode);
            xmlDocument.AppendChild(xmlRoot);
            NotificationManager.Notify(Notifications.ProductInfo.OnAfterGenerateProductInfoXml, new Notifications.ProductInfo.OnAfterGenerateProductInfoXmlArgs(products, settings, xmlDocument));
            if (settings.Beautify)
            {
                return xmlDocument.Beautify();
            }
            else
            {
                return xmlDocument.InnerXml;
            }
        }

        /// <summary>
        /// Builds the product information XML.
        /// </summary>
        /// <param name="xmlDocument">The XML document.</param>
        /// <param name="products">The products.</param>
        /// <param name="settings">The settings.</param>
        /// <returns>XmlNode.</returns>
        private XmlNode BuildProductInfoXml(XmlDocument xmlDocument, Dictionary<Product, double> products, ProductInfoXmlGeneratorSettings settings)
        {
            if (products == null || !products.Any())
            {
                return null;
            }

            var currencyCode = Common.Context.Currency.Code;

            var tableNode = xmlDocument.CreateElement("Products");
            tableNode.SetAttribute("type", "filter");

            bool doCalculatePriceUsingProductNumber = Settings.Instance.CalculatePriceUsingProductNumber;

            foreach (var productWithQuantity in products)
            {
                Product product = productWithQuantity.Key;
                var itemNode = xmlDocument.CreateElement("Product");
                tableNode.AppendChild(itemNode);

                if (product != null)
                {
                    AddChildXmlNode(itemNode, "ProductId", doCalculatePriceUsingProductNumber ? product.Number : product.Id);
                    AddChildXmlNode(itemNode, "ProductVariantId", product.VariantId);
                    AddChildXmlNode(itemNode, "ProductNumber", product.Number);
                    AddChildXmlNode(itemNode, "ProductIdentifier", ProductManager.ProductProvider.GetProductIdentifier(product));
                    AddChildXmlNode(itemNode, "CurrencyCode", currencyCode);
                    AddChildXmlNode(itemNode, "Quantity", productWithQuantity.Value.ToIntegrationString());

                    if (settings.AddProductFieldsToRequest && product.ProductFieldValues.Count > 0)
                    {
                        AppendProductFields(product, itemNode);
                    }
                }
            }

            return tableNode;
        }

        /// <summary>
        /// Appends the product fields.
        /// </summary>
        /// <param name="product">The product.</param>
        /// <param name="productNode">The product node.</param>
        private void AppendProductFields(Product product, XmlElement productNode)
        {
            foreach (var field in product.ProductFieldValues)
            {
                string value = field.Value?.ToIntegrationString() ?? string.Empty;
                AddChildXmlNode(productNode, field.ProductField.SystemName, value, isCustomField: true);
            }
        }

        private void AddChildXmlNode(XmlElement parent, string nodeName, string nodeValue, bool isCustomField = false)
        {
            if (parent.OwnerDocument == null)
            {
                throw new ArgumentException("parent element must have an OwnerDocument", nameof(parent));
            }

            var node = parent.OwnerDocument.CreateElement(nodeName);
            if (isCustomField)
            {
                node.SetAttribute("isCustomField", true.ToString());
            }

            node.InnerText = nodeValue;
            parent.AppendChild(node);
        }
    }
}